/*########################################################################
	File:			AAFractalImage.cpp
	Creation date:	25th May 2002

	Copyright 2002 Mark Williams, all rights reserved.
	Contact: mark@extrabit.com
	http://www.extrabit.com

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

########################################################################*/

/*########################################################################
	Headers
########################################################################*/

#include "StdAfx.h"

#include "Processor.h"
#include "Palette.h"
#include "FractalImage.h"
#include "AAFractalImage.h"

/*########################################################################
	Local declarations
########################################################################*/

/*########################################################################
	Local definitions
########################################################################*/


/**	Constructor.
 *	@param image The image that this AAFractalImage will anti-alias.
 */
AAFractalImage::AAFractalImage (FractalImage* image)
{
	m_image = image ;
	m_subpixel_gridsize = 4 ;
	m_done = false ;
	m_progress = 0 ;
	m_x = 0 ;
	m_y = 0 ;
	m_curr_rgb_pixel = 0 ;
}


/**	Sets the sub pixel grid size.
 *	@param subpixel_gridsize Sets the size of the sub-pixel grid used to anti-alias.
 *	The number set defines the size of the square grid, so
 *	there will be the square of this number of sub-pixels.
 *	If 'n'=1, then there is only one sample per pixel, i.e. no anti-aliasing.
 *	If 'n'=2, then 2x2 (i.e. 4) samples are calculated per pixel.
 *	If 'n'=3, then 3x3 (i.e. 9) samples, and so on.
 */
void AAFractalImage::set_subpixel_gridsize (int subpixel_gridsize)
{
	m_subpixel_gridsize = subpixel_gridsize ;
}


/**	Initialises all the member variables for anti-aliasing 'm_image'.
 *	This must be called before process() is called, and whenever
 *	the image or sub pixel grid size changes.
 */
void AAFractalImage::restart (void)
{
	// Calculate 'm_xoffs' and 'm_yoffs'.
	Vector2 org   (0.0, 0.0) ;
	Vector2 org_x (1.0, 0.0) ;
	Vector2 org_y (0.0, 1.0) ;
	const Matrix33& view_transform = m_image->inq_view_transform() ;
	view_transform.transform (&org) ;
	view_transform.transform (&org_x) ;
	view_transform.transform (&org_y) ;
	m_xoffs = org_x - org ;
	m_yoffs = org_y - org ;
	// Calculate subpixel offsets.
	m_xsuboffs = m_xoffs / (double)m_subpixel_gridsize ;
	m_ysuboffs = m_yoffs / (double)m_subpixel_gridsize ;

	// Set 'm_rowstart' to the centre of the first sub pixel in the first (bottom) row.
	double suboffs = 0.5/(double)m_subpixel_gridsize ;
	m_rowstart = Vector2 (suboffs, suboffs) ;
	view_transform.transform (&m_rowstart) ;

	m_curr_pixel = m_rowstart ;
	m_curr_rgb_pixel = m_image->inq_rgb_pixels() ;

	m_done = false ;
	m_progress = 0 ;
	m_x = 0 ;
	m_y = 0 ;
}


/**	@return Returns the integer closest to d
 */
inline int roundoff (double d)
{
#ifdef _X86_
	// The default processor rounding mode is round off,
	// so this implementation uses the fistp instruction directly.
	int retval ;
	_asm
	{
		fld d
		fistp retval
	}
	return retval ;
#else
	// Rounding in C++ is round to zero, so need to use floor() function
	// so that roundoff works correctly with negative numbers.
	return (int)(floor(d+0.5)) ;
#endif
}


/**	Processes the next pixel.
 *	This method should be called repeatedly until inq_done() returns true.
 *	@return Returns true if the display should be updated, otherwise false.
 */
bool AAFractalImage::process (void)
{
	if (m_done) return false ;

	int width = m_image->inq_width() ;
	int height = m_image->inq_height() ;

	// Check if current pixel needs to be processed.
	// Only anti-alias pixels that are different to any of their neighbours.
	bool do_process = false ;
	if (m_x > 0 && m_y > 0 && m_x < width-1 && m_y < height-1)
	{
		const unsigned short* pixel_itns = m_image->inq_pixel_itns (m_x, m_y) ;
		int itns = *pixel_itns ;
		if (itns != pixel_itns[-1]) do_process = true ;
		else if (itns != pixel_itns[1]) do_process = true ;
		else if (itns != pixel_itns[-width]) do_process = true ;
		else if (itns != pixel_itns[width]) do_process = true ;
		else if (itns != pixel_itns[-1-width]) do_process = true ;
		else if (itns != pixel_itns[-1+width]) do_process = true ;
		else if (itns != pixel_itns[1-width]) do_process = true ;
		else if (itns != pixel_itns[1+width]) do_process = true ;
	}
	else
	{
		// Anti-alias edge pixels anway.
		do_process = true ;
	}

	if (do_process)
	{
		const Processor* processor = m_image->inq_processor() ;
		ComplexNumber initial_z = m_image->inq_initial_z() ;
		const Palette* palette = m_image->inq_palette() ;
		const float* colours = palette->inq_linear_colours() ;
		unsigned short max_itns = (unsigned short)(palette->inq_ncolours() - 1) ;
		double r = 0 ;
		double g = 0 ;
		double b = 0 ;
		Vector2 vrow = m_curr_pixel ;
		for (int i = 0 ; i < m_subpixel_gridsize ; i++)
		{
			Vector2 v = vrow ;
			for (int j = 0 ; j < m_subpixel_gridsize ; j++)
			{
				ComplexNumber z = m_image->convert_point_to_z (v) ;
				unsigned short itns = processor->iterate_to_end (initial_z, z, max_itns) ;
				const float* col = &colours[3*itns] ;
				r += col[0] ;
				g += col[1] ;
				b += col[2] ;
				v += m_xsuboffs ;
			}
			vrow += m_ysuboffs ;
		}

		int n2 = m_subpixel_gridsize*m_subpixel_gridsize ;
		r /= n2 ;
		g /= n2 ;
		b /= n2 ;

		r = palette->gamma_correct (r) ;
		g = palette->gamma_correct (g) ;
		b = palette->gamma_correct (b) ;

		rgbcolour_t pixel = (roundoff (r*255.0) << 16) | (roundoff (g*255.0) << 8) | roundoff (b*255.0) ;
		*m_curr_rgb_pixel++ = pixel ;
	}
	else
	{
		m_curr_rgb_pixel++ ;
	}

	m_curr_pixel += m_xoffs ;
	if (++m_x == width)
	{
		m_x = 0 ;
		m_rowstart += m_yoffs ;
		m_curr_pixel = m_rowstart ;
		if (++m_y == height)
		{
			m_done = true ;
		}

		// Update progress counter - just based on current line.
		m_progress = (int)(100.0*(double)m_y/(double)m_image->inq_height()) ;
	}

	return m_done ;
}


/*########################################################################
	End
########################################################################*/
