/*########################################################################
	File:			FractalImage.cpp
	Creation date:	25th May 2002

	Copyright 2002 Mark Williams, all rights reserved.
	Contact: mark@extrabit.com
	http://www.extrabit.com

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

########################################################################*/

/*########################################################################
	Headers
########################################################################*/

#include "StdAfx.h"

#include "Palette.h"
#include "FractalImage.h"

/*########################################################################
	Local declarations
########################################################################*/

/*########################################################################
	Local definitions
########################################################################*/


/**	Constructor.
 */
FractalImage::FractalImage (void)
{
	m_processor = 0 ;
	m_invert_u = false ;
	m_width = 0 ;
	m_height = 0 ;
	m_pixel_itns = 0 ;
	m_rgb_pixels = 0 ;
	m_view_width = 0.0 ;
	m_view_angle = 0.0 ;
}


/**	Copy constructor.
 *	Copies the parameters from the given FractalImage.
 *	Note: any pixel values already processed in the given FractalImage are not copied.
 *	resize()/init() must be called before this FractalImage is used.
 */
FractalImage::FractalImage (const FractalImage& image)
{
	m_processor = image.m_processor ;
	m_invert_u  = image.m_invert_u ;
	m_initial_z = image.m_initial_z ;
	m_palette = image.m_palette ;
	m_width   = image.m_width ;
	m_height  = image.m_height ;
	m_pixel_itns = 0 ;
	m_rgb_pixels = 0 ;
	m_view_centre = image.m_view_centre ;
	m_view_width  = image.m_view_width ;
	m_view_angle  = image.m_view_angle ;
	m_view_transform = image.m_view_transform ;
}


/**	Destructor.
 */
FractalImage::~FractalImage (void)
{
	delete[] m_pixel_itns ;
}


/**	Sets the Processor to use.
 *	This must be set before any other methods are called.
 */
void FractalImage::set_processor (const Processor* processor)
{
	m_processor = processor ;
}


/**	Sets the Palette to use.
 *	This must be set before any other methods are called.
 */
void FractalImage::set_palette (const Palette* palette)
{
	m_palette = palette ;
}


/**	Resizes the image to the given width & height in pixels.
 *	set_view_transform() must be called after this method.
 */
void FractalImage::resize (int width, int height, rgbcolour_t* rgb_pixels)
{
	m_width = width ;
	m_height = height ;

	delete[] m_pixel_itns ;
	m_pixel_itns = new unsigned short [m_width*m_height] ;

	m_rgb_pixels = rgb_pixels ;

	update_view_transform() ;
}


/**	Sets the view transform.
 *	@param centre Point in fractal space on which view is centred.
 *	@param width Width of view in fractal space.
 *	@param angle Anti-clockwise angle of rotation of view (radians), about 'centre'.
 */
void FractalImage::set_view (const Vector2& centre, double width, double angle)
{
	m_view_centre = centre ;
	m_view_width = width ;
	m_view_angle = angle ;

	if (m_width > 0)
	{
		update_view_transform() ;
	}
}


/**	This should be called whenever a new image is being calculated.
 *	It sets all the pixel values to the main colour.
 */
void FractalImage::restart (void)
{
	unsigned short max_itn = (unsigned short)(m_palette->inq_ncolours()-1) ;
	rgbcolour_t main_colour = m_palette->inq_main_colour() ;

	int n = m_width*m_height ;
	for (int i = 0 ; i < n ; i++)
	{
		m_pixel_itns[i] = max_itn ;
		m_rgb_pixels[i] = main_colour ;
	}
}


/**	Updates the colours of all RGB pixels.
 *	This is called whenever the palette is changed.
 */
void FractalImage::update_all_rgb_pixels (void)
{
	int n = inq_npixels() ;
	const rgbcolour_t* colours = m_palette->inq_colours() ;
	for (int i = 0 ; i < n ; i++)
	{
		unsigned short n_itns = m_pixel_itns[i] ;
		m_rgb_pixels[i] = colours[n_itns] ;
	}
}


/**	Converts the given point into a complex number.
 *	Normally, this simply involves copying the x-coordinate to the
 *	real component, and the y-coordinate to the imaginary component,
 *	but other transforms such as taking the recipricol or square root
 *	may be applied to apply distortions to the image.
 */
ComplexNumber FractalImage::convert_point_to_z (const Vector2& v) const
{
	ComplexNumber rtn (v.x(), v.y()) ;
	if (m_invert_u)
	{
		rtn = rtn.inq_inverse() ;
	}
	return rtn ;
}


/**	Updates the view transform.
 *	This should be called whenever the view or image size is changed.
 */
void FractalImage::update_view_transform (void)
{
	double scale = m_view_width / (double)m_width ;

	// Build up transform by combining:
	//	translate centre of image to origin
	//	scale about origin
	//	rotate about origin
	//	translate origin to view centre
	Matrix33 translate1_transform (Matrix33::TRANSLATE, Vector2 (-0.5*m_width, -0.5*m_height)) ;
	Matrix33 scale_transform (Matrix33::SCALE, scale) ;
	Matrix33 rotate_transform (Matrix33::ROTATE, m_view_angle) ;
	Matrix33 translate2_transform (Matrix33::TRANSLATE, m_view_centre) ;
	m_view_transform = translate2_transform * rotate_transform * scale_transform * translate1_transform ;
}


/*########################################################################
	End
########################################################################*/
