#ifndef FractalImage_H
#define FractalImage_H
/*########################################################################
	File:			FractalImage.h
	Creation date:	25th May 2002

	Copyright 2002 Mark Williams, all rights reserved.
	Contact: mark@extrabit.com
	http://www.extrabit.com

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

########################################################################*/

/*########################################################################
	Headers
########################################################################*/

#include "rgbcolour.h"
#include "Vector2.h"
#include "ComplexNumber.h"
#include "Matrix33.h"

/*########################################################################
	Declarations
########################################################################*/

class Processor ;
class Palette ;

/*########################################################################
	Class Definition
########################################################################*/

/**	FractalImage stores all the parameters for a fractal image being
 *	generated, and the computed results.
 *	This class stores both the number of iterations calculated for each
 *	pixel and the RGB colour for each pixel.  This allows the RGB colours
 *	to be quickly recalculated if the palette is changec.
 *	This class also stores a pointer to the pixel processor 'm_processor'
 *	which performs the fractal calculations at each pixel.
 *	This is used by the FastImageGen and ProgressiveImageGen classes to
 *	generate this image.
 */
class FractalImage
{
public:
	FractalImage (void) ;
	FractalImage (const FractalImage&) ;
	~FractalImage (void) ;

	const Processor* inq_processor (void) const { return m_processor ; }
	void set_processor (const Processor*) ;
	const ComplexNumber& inq_initial_z (void) const { return m_initial_z ; }
	void set_initial_z (const ComplexNumber& z) { m_initial_z = z ; }
	void set_invert_u (bool invert_u) { m_invert_u = invert_u ; }
	bool inq_invert_u (void) const { return m_invert_u ; }

	void set_palette (const Palette*) ;
	const Palette* inq_palette (void) const { return m_palette ; }
	void resize (int width, int height, rgbcolour_t* rgb_pixels) ;

	void set_view (const Vector2& centre, double width, double angle) ;
	const Matrix33& inq_view_transform (void) const { return m_view_transform ; }
	const Vector2& inq_view_centre (void) const { return m_view_centre ; }
	double inq_view_width (void) const { return m_view_width ; }
	double inq_view_angle (void) const { return m_view_angle ; }

	int inq_width (void) const { return m_width ; }
	int inq_height (void) const { return m_height ; }
	int inq_npixels (void) const { return m_width*m_height ; }
	unsigned short* inq_pixel_itns (void) const { return m_pixel_itns ; }
	const unsigned short* inq_pixel_itns (int x, int y) const { return m_pixel_itns + x + y*m_width ; }
	rgbcolour_t* inq_rgb_pixels (void) const { return m_rgb_pixels ; }

	void restart (void) ;
	void update_all_rgb_pixels (void) ;

	ComplexNumber convert_point_to_z (const Vector2&) const ;

private:
	void update_view_transform (void) ;

private:
	/**	Fractal image generator. */
	const Processor* m_processor ;

	/**	If set, u is set to the reciprical of the coordinate before
	 *	calculating the fractal at that point.
	 */
	bool m_invert_u ;

	/** Starting value for 'z'. */
	ComplexNumber m_initial_z ;

	/**	Palette to use. */
	const Palette* m_palette ;

	/** Width in pixels. */
	int m_width ;

	/** Height in pixels. */
	int m_height ;

	/** Pointer to array of 'width' x 'height' unsigned shorts.
	 *	As the iterated value at each pixel exceeds the divergence
	 *	threshold, the iteration count is set here for each pixel.
	 *	Lines are order bottom to top, as is conventional for
	 *	windows bitmaps.
	 *	Note: this is the reverse to the convention for the coordinate
	 *	space used by windows, which have (0,0) at top left.
	 */
	unsigned short* m_pixel_itns ;

	/** Pointer to RGB pixels.
	 *	There are 'm_width' x 'm_height' pixels, ordered bottom to top
	 *	row by row.
	 *	These pixels are owned by the HBITMAP in CFracWnd, so it is that
	 *	class's responsibility to make sure that if these pixels are
	 *	invalidated (such as when the window is resized) this class is
	 *	notified.
	 */
	rgbcolour_t* m_rgb_pixels ;

	/**	Centre of view in fractal space. */
	Vector2 m_view_centre ;

	/**	Width of view in fractal space. */
	double m_view_width ;

	/**	Angle in radians that view is rotated about view centre. */
	double m_view_angle ;

	/** View transform.
	 *	Transforms a coordinate in image pixel space (y=0 at bottom)
	 *	to the fractal space.
	 */
	Matrix33 m_view_transform ;
} ;


/*########################################################################
	End
########################################################################*/
#endif
