/*########################################################################
	FractalViewer.cpp

	Copyright 2002 Mark Williams, all rights reserved.
	Contact: mark@extrabit.com
	http://www.extrabit.com

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

########################################################################*/

#include "stdafx.h"
#include "FractalViewer.h"


#include "MandelbrotProcessor.h"
#include "Z3Processor.h"
#include "Z4Processor.h"
#include "Z5Processor.h"
#include "Z6Processor.h"
#include "Z7Processor.h"

#include "MainFrm.h"
#include "OptionsDlg.h"
#include "FractalDlg.h"
#include "ExportImageDlg.h"
#include "ProgressDlg.h"
#include "JpgWriter.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CFractalViewerApp

BEGIN_MESSAGE_MAP(CFractalViewerApp, CWinApp)
	//{{AFX_MSG_MAP(CFractalViewerApp)
	ON_COMMAND(ID_APP_ABOUT, OnAppAbout)
	ON_COMMAND(ID_VIEW_OPTIONS, OnViewOptions)
	ON_COMMAND(ID_PALETTE_RANDOMISE, OnPaletteRandomise)
	ON_COMMAND(ID_VIEW_FRACTALS, OnViewFractals)
	ON_COMMAND(ID_PALETTE_MAIN_COL, OnPaletteMainCol)
	ON_COMMAND(ID_PALETTE_BRIGHT, OnPaletteBright)
	ON_COMMAND(ID_PALETTE_PASTEL, OnPalettePastel)
	ON_COMMAND(ID_PALETTE_MONOCHROME, OnPaletteMonochrome)
	ON_COMMAND(ID_PALETTE_WHITE, OnPaletteWhite)
	ON_UPDATE_COMMAND_UI(ID_PALETTE_BRIGHT, OnUpdatePaletteBright)
	ON_UPDATE_COMMAND_UI(ID_PALETTE_PASTEL, OnUpdatePalettePastel)
	ON_UPDATE_COMMAND_UI(ID_PALETTE_MONOCHROME, OnUpdatePaletteMonochrome)
	ON_UPDATE_COMMAND_UI(ID_PALETTE_WHITE, OnUpdatePaletteWhite)
	ON_COMMAND(ID_EXPORT_JPG, OnExportJpg)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


static const int DEFAULT_PROCESSOR_TYPE = 0 ;
static const int DEFAULT_PALETTE_SIZE = 256 ;

static const double DEFAULT_CENTRE_X = 0.0 ;
static const double DEFAULT_CENTRE_Y = 0.0 ;
static const double DEFAULT_WIDTH = 4.0 ;
static const double DEFAULT_ANGLE = 0.0 ;


/////////////////////////////////////////////////////////////////////////////
// CFractalViewerApp construction

CFractalViewerApp::CFractalViewerApp() :
	m_palette (DEFAULT_PALETTE_SIZE),
	m_progressive_imagegen (&m_image),
	m_fast_imagegen (&m_image),
	m_aa_image (&m_image)
{
	set_processor (DEFAULT_PROCESSOR_TYPE) ;

	m_palette.randomise() ;
	m_image.set_palette (&m_palette) ;

	m_progressive_imagegen.init() ;
	m_fast_imagegen.init() ;

	m_progressive = true ;
	set_imagegen() ;

	m_imagegen = &m_progressive_imagegen ;

	CFractalDlg::set_default_view (DEFAULT_CENTRE_X, DEFAULT_CENTRE_Y,
			DEFAULT_WIDTH, DEFAULT_ANGLE) ;

	m_image.set_view (Vector2 (DEFAULT_CENTRE_X, DEFAULT_CENTRE_Y),
			DEFAULT_WIDTH, DEFAULT_ANGLE) ;

	m_progress = -1 ;
}


/////////////////////////////////////////////////////////////////////////////
// The one and only CFractalViewerApp object

CFractalViewerApp theApp;


/**	This is called when the FractalImage has been resized.
 *	The ProgressiveFractalImage is re-initialised.
 */
void CFractalViewerApp::image_resized (void)
{
	m_progressive_imagegen.init() ;
	m_fast_imagegen.init() ;
	restart() ;
}


/** This is called when a change is made to the FractalImage.
 *	The ProgressiveFractalImage is restarted.
 */
void CFractalViewerApp::restart (void)
{
	m_image.restart() ;
	m_imagegen->restart() ;
}


/**	Updates the status bar to show the current view.
 *	This should be called whenever the view has been changed.
 */
void CFractalViewerApp::update_status_bar (void)
{
	double x = m_image.inq_view_centre().x() ;
	double y = m_image.inq_view_centre().y() ;
	double width = m_image.inq_view_width() ;
	double angle = m_image.inq_view_angle() ;

	((CMainFrame*)AfxGetMainWnd())->update_status_bar (x, y, width, angle) ;
}


/**	Sets the fractal type to that given.
 *	The different fractal types are implemented by different derived
 *	Processor classes.
 */
void CFractalViewerApp::set_processor (int fractal_type)
{
	static MandelbrotProcessor mandelbrot_processor ;
	static Z3Processor z3_processor ;
	static Z4Processor z4_processor ;
	static Z5Processor z5_processor ;
	static Z6Processor z6_processor ;
	static Z7Processor z7_processor ;

	m_fractal_type = fractal_type ;

	// Sets the processor in 'm_image', depending on the value of 'm_fractal_type'.
	switch (m_fractal_type)
	{
	case 0:	m_image.set_processor (&mandelbrot_processor) ; break ;
	case 1:	m_image.set_processor (&z3_processor) ; break ;
	case 2:	m_image.set_processor (&z4_processor) ; break ;
	case 3:	m_image.set_processor (&z5_processor) ; break ;
	case 4:	m_image.set_processor (&z6_processor) ; break ;
	case 5:	m_image.set_processor (&z7_processor) ; break ;
	}
}


/**	Sets 'm_imagegen' to point to either 'm_progressive_imagegen' or
 *	'm_fast_imagegen', depending on the value of 'm_progressive'.
 */
void CFractalViewerApp::set_imagegen (void)
{
	m_imagegen = m_progressive ? (ImageGen*)(&m_progressive_imagegen) : (ImageGen*)(&m_fast_imagegen) ;
}


/**	Updates the palette, regenerating the colours it contains.
 *	This is called to re-randomise the colours in the palette,
 *	or if the type is changed.
 */
void CFractalViewerApp::update_palette (void)
{
	// Randomise the palette.
	m_palette.randomise() ;

	// Update the display.
	m_image.update_all_rgb_pixels() ;
	((CMainFrame*)m_pMainWnd)->update_frac_wnd() ;

	// Reset AA image since anti-aliasing will need to be redone.
	m_aa_image.restart() ;
}


/**	Creates an image of the given size, and saves it to the given path.
 */
void CFractalViewerApp::export_image (CString path, int width, int height, int aa_gridsize)
{
	FractalImage image (m_image) ;
	rgbcolour_t* rgb_pixels = new rgbcolour_t[width*height] ;
	image.resize (width, height, rgb_pixels) ;
	image.restart() ;

	FastImageGen gen (&image) ;
	gen.init() ;
	gen.restart() ;

	// Disable UI while progress panel is shown.
	m_pMainWnd->EnableWindow (FALSE) ;

	CProgressDlg dlg ;
	dlg.Create (IDD_PROGRESS_DLG) ;
	dlg.set_text (IDS_GENERATING_IMAGE) ;

	int progress = -1 ;
	while (!gen.inq_done())
	{
		if (dlg.inq_abort()) break ;

		gen.process() ;
		int p = gen.inq_progress() ;
		if (p != progress)
		{
			progress = p ;
			dlg.set_progress (progress) ;
		}
	}

	if (!dlg.inq_abort() && aa_gridsize > 1)
	{
		AAFractalImage aa_image (&image) ;
		aa_image.set_subpixel_gridsize (aa_gridsize) ;
		aa_image.restart() ;

		dlg.set_text (IDS_ANTIALIASING_IMAGE) ;
		while (!aa_image.inq_done())
		{
			if (dlg.inq_abort()) break ;

			aa_image.process() ;
			int p = aa_image.inq_progress() ;
			if (p != progress)
			{
				progress = p ;
				dlg.set_progress (progress) ;
			}
		}
	}

	if (!dlg.inq_abort())
	{
		dlg.set_text (IDS_SAVING_IMAGE) ;
		dlg.inq_abort() ;	// Run event loop to update text.

		const int jpg_quality = 100 ;		// Maximum quality.
		JpgWriter writer (jpg_quality) ;
		try
		{
			writer.write_image (path, width, height, rgb_pixels) ;
		}
		catch (CString str)
		{
			AfxMessageBox (str) ;
		}
	}

	// Re-enable UI.
	m_pMainWnd->EnableWindow (TRUE) ;

	dlg.DestroyWindow() ;

	delete[] rgb_pixels ;
}


/////////////////////////////////////////////////////////////////////////////
// CFractalViewerApp initialization

BOOL CFractalViewerApp::InitInstance()
{
	// Standard initialization
	// If you are not using these features and wish to reduce the size
	//  of your final executable, you should remove from the following
	//  the specific initialization routines you do not need.

#ifdef _AFXDLL
	Enable3dControls();			// Call this when using MFC in a shared DLL
#else
	Enable3dControlsStatic();	// Call this when linking to MFC statically
#endif

	SetRegistryKey(_T("ExtraBit Software"));


	// To create the main window, this code creates a new frame window
	// object and then sets it as the application's main window object.

	CMainFrame* pFrame = new CMainFrame;
	m_pMainWnd = pFrame;

	// create and load the frame with its resources

	pFrame->LoadFrame(IDR_MAINFRAME,
		WS_OVERLAPPEDWINDOW, NULL,
		NULL);

	// The one and only window has been initialized, so show and update it.
	pFrame->ShowWindow(SW_SHOW);
	pFrame->UpdateWindow();

	update_status_bar() ;

	return TRUE;
}


/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
	afx_msg void OnAboutImage();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
	ON_BN_CLICKED(IDC_ABOUT_IMAGE, OnAboutImage)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

// App command to run the dialog
void CFractalViewerApp::OnAppAbout()
{
	CAboutDlg aboutDlg;
	aboutDlg.DoModal();
}

/////////////////////////////////////////////////////////////////////////////
void CAboutDlg::OnAboutImage() 
{
	ShellExecute (m_hWnd, _T("open"), _T("http://www.extrabit.com"), NULL, NULL, SW_SHOW) ;
}

/////////////////////////////////////////////////////////////////////////////
// CFractalViewerApp message handlers


/////////////////////////////////////////////////////////////////////////////
BOOL CFractalViewerApp::OnIdle(LONG lCount) 
{
	if (CWinApp::OnIdle(lCount))
	{
		// Let Windows do all it's idle processing first.
		return TRUE ;
	}

	bool all_done = false ;

	if (!m_imagegen->inq_done())
	{
		// Generate aliased image first.
		if (m_imagegen->process())
		{
			((CMainFrame*)m_pMainWnd)->update_frac_wnd() ;
		}

		int progress = m_imagegen->inq_progress() ;
		if (m_progress != progress)
		{
			m_progress = progress ;

			// Set status bar message to show progress.
			CString msg ;
			int ids = m_progressive ? IDS_SHOW_ITN : IDS_GENERATING ;
			msg.Format (ids, m_progress) ;
			((CMainFrame*)m_pMainWnd)->show_status (msg) ;
		}

		if (m_imagegen->inq_done())
		{
			// Next time, AA image will be processed.
			m_aa_image.restart() ;
		}
	}
	else
	{
		// If aliased image generated, perform anti-aliasing
		if (m_aa_image.process())
		{
			((CMainFrame*)m_pMainWnd)->update_frac_wnd() ;
		}

		int progress = m_aa_image.inq_progress() ;
		if (m_progress != progress)
		{
			m_progress = progress ;

			// Set status bar message to show progress.
			CString msg ;
			msg.Format (IDS_ANTIALIASING, m_progress) ;
			((CMainFrame*)m_pMainWnd)->show_status (msg) ;
		}

		all_done = m_aa_image.inq_done() ;
		if (all_done)
		{
			// Set status bar message to show processing finished.
			CString msg ;
			msg.Format (AFX_IDS_IDLEMESSAGE) ;
			((CMainFrame*)m_pMainWnd)->show_status (msg) ;
		}
	}

	// Needs to return TRUE if OnIdle() should be called again,
	// i.e. if there is any more processing to do.
	return all_done ? FALSE : TRUE ;
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnViewOptions() 
{
	int aa_gridsize = m_aa_image.inq_subpixel_gridsize() ;
	int palette_size = m_palette.inq_ncolours() ;

	COptionsDlg	dlg ;
	dlg.m_palette_size = palette_size ;
	dlg.set_aa_gridsize (aa_gridsize) ;
	dlg.m_progressive = m_progressive ? TRUE : FALSE ;
	if (dlg.DoModal() == IDOK)
	{
		if (aa_gridsize != dlg.inq_aa_gridsize())
		{
			m_aa_image.set_subpixel_gridsize (dlg.inq_aa_gridsize()) ;
			m_aa_image.restart() ;
		}

		if (palette_size != dlg.m_palette_size)
		{
			palette_size = dlg.m_palette_size ;
			m_palette.resize (palette_size) ;

			m_palette.randomise() ;
			restart() ;
		}

		bool progressive = (dlg.m_progressive != FALSE) ;
		if (m_progressive != progressive)
		{
			m_progressive = progressive ;
			set_imagegen() ;
			restart() ;
		}
	}
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnViewFractals() 
{
	const double PI = 3.141592653589793238462643383279502884197 ;
	const double DEGREES_TO_RADIANS = 180.0 / PI ;

	double view_x = m_image.inq_view_centre().x() ;
	double view_y = m_image.inq_view_centre().y() ;
	double view_width = m_image.inq_view_width() ;
	double view_angle_degrees = m_image.inq_view_angle() * DEGREES_TO_RADIANS ;

	CFractalDlg dlg ;
	dlg.set_fractal_type (m_fractal_type) ;
	dlg.m_invert_u = m_image.inq_invert_u() ? TRUE : FALSE ;
	dlg.m_zx = m_image.inq_initial_z().x() ;
	dlg.m_zy = m_image.inq_initial_z().y() ;
	dlg.m_view_x = view_x ;
	dlg.m_view_y = view_y ;
	dlg.m_view_width = view_width ;
	dlg.m_view_angle = view_angle_degrees ;

	if (dlg.DoModal() == IDOK)
	{
		// Set fractal type.
		set_processor (dlg.inq_fractal_type()) ;

		// Set invert u
		bool invert_u = (dlg.m_invert_u != FALSE) ;
		m_image.set_invert_u (invert_u) ;
		// Set initial 'z'
		ComplexNumber initial_z (dlg.m_zx, dlg.m_zy) ;
		m_image.set_initial_z (initial_z) ;

		// Set view - this automatically restarts the image generation.
		m_image.set_view (Vector2 (dlg.m_view_x, dlg.m_view_y), dlg.m_view_width, dlg.m_view_angle/DEGREES_TO_RADIANS) ;
		restart() ;
		update_status_bar() ;
	}
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnPaletteRandomise() 
{
	update_palette() ;
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnPaletteMainCol() 
{
	CColorDialog dlg ;

	rgbcolour_t pal_col = m_palette.inq_main_colour() ;
	BYTE r = (BYTE)((pal_col >> 16) & 0xFF) ;
	BYTE g = (BYTE)((pal_col >> 8) & 0xFF) ;
	BYTE b = (BYTE)(pal_col & 0xFF) ;
	dlg.m_cc.rgbResult = RGB(r,g,b) ;
	dlg.m_cc.Flags = CC_RGBINIT|CC_ENABLEHOOK ;

	if (dlg.DoModal() == IDOK)
	{
		COLORREF dlg_col = dlg.GetColor() ;
		r = GetRValue(dlg_col) ;
		g = GetGValue(dlg_col) ;
		b = GetBValue(dlg_col) ;
		m_palette.set_main_colour ((r << 16) | (g << 8) | b) ;

		// Update the display.
		m_image.update_all_rgb_pixels() ;
		((CMainFrame*)m_pMainWnd)->update_frac_wnd() ;

		// Reset AA image since anti-aliasing will need to be redone.
		m_aa_image.restart() ;
	}
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnPaletteBright() 
{
	if (m_palette.inq_type() != PALETTE_BRIGHT)
	{
		m_palette.set_type (PALETTE_BRIGHT) ;
		update_palette() ;
	}
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnPalettePastel() 
{
	if (m_palette.inq_type() != PALETTE_PASTEL)
	{
		m_palette.set_type (PALETTE_PASTEL) ;
		update_palette() ;
	}
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnPaletteMonochrome() 
{
	if (m_palette.inq_type() != PALETTE_MONOCHROME)
	{
		m_palette.set_type (PALETTE_MONOCHROME) ;
		update_palette() ;
	}
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnPaletteWhite() 
{
	if (m_palette.inq_type() != PALETTE_WHITE)
	{
		m_palette.set_type (PALETTE_WHITE) ;
		update_palette() ;
	}
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnUpdatePaletteBright(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck ((m_palette.inq_type() == PALETTE_BRIGHT) ? 1 : 0) ;
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnUpdatePalettePastel(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck ((m_palette.inq_type() == PALETTE_PASTEL) ? 1 : 0) ;
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnUpdatePaletteMonochrome(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck ((m_palette.inq_type() == PALETTE_MONOCHROME) ? 1 : 0) ;
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnUpdatePaletteWhite(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck ((m_palette.inq_type() == PALETTE_WHITE) ? 1 : 0) ;
}


/////////////////////////////////////////////////////////////////////////////
void CFractalViewerApp::OnExportJpg() 
{
	// Store width & height in static variables so values set will be remembered
	// next time.
	static int image_width  = 640 ;
	static int image_height = 480 ;
	static int aa_gridsize  = 4 ;

	CExportImageDlg export_dlg ;
	export_dlg.m_width = image_width ;
	export_dlg.m_height = image_height ;
	export_dlg.set_aa_gridsize (aa_gridsize) ;
	if (export_dlg.DoModal() == IDOK)
	{
		image_width  = export_dlg.m_width ;
		image_height = export_dlg.m_height ;
		aa_gridsize  = export_dlg.inq_aa_gridsize() ;

		CFileDialog save_dlg (FALSE, "jpg", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT, "Jpg Image (*.jpg)|*.jpg||") ;
		if (save_dlg.DoModal() == IDOK)
		{
			CString path = save_dlg.GetPathName() ;
			export_image (path, image_width, image_height, aa_gridsize) ;
		}
	}
}
