/*########################################################################
	File:			JpgWriter.cpp
	Creation date:	6th June 2002

	Copyright 2002 Mark Williams, all rights reserved.
	Contact: mark@extrabit.com
	http://www.extrabit.com

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

########################################################################*/

/*########################################################################
	Headers
########################################################################*/

#include "StdAfx.h"

extern "C"
{
namespace jpg
{
#include <jpeglib.h>
}
}

#include "JpgWriter.h"

/*########################################################################
	Local declarations
########################################################################*/

/*########################################################################
	Local definitions
########################################################################*/


/**	A pointer to this function is set in the struct jpeg_error_mgr which is
 *	used by JpgWriter::write_image().
 *	It formats the error message, and throws a CString exception.
 */
static void jpeg_error_exit_handler (jpg::j_common_ptr cinfo)
{
	char buffer[JMSG_LENGTH_MAX];

	/* Create the message */
	(*cinfo->err->format_message) (cinfo, buffer);

	throw CString (buffer) ;
}


/**	Constructor.
 */
JpgWriter::JpgWriter (int quality)
{
	m_quality = quality ;
}


/**	Writes the given image to the given file.
 *	Throws a CString exception containing the error message if there is
 *	a problem writing the file.
 */
void JpgWriter::write_image (CString path, int width, int height, const rgbcolour_t* rgb_data)
{
	/* This struct contains the JPEG compression parameters and pointers to
	* working space (which is allocated as needed by the JPEG library).
	* It is possible to have several such structures, representing multiple
	* compression/decompression processes, in existence at once.  We refer
	* to any one struct (and its associated working data) as a "JPEG object".
	*/
	struct jpg::jpeg_compress_struct cinfo;

	struct jpg::jpeg_error_mgr jerr;
	/* We set up the normal JPEG error routines, then override error_exit. */
	jerr.error_exit = jpeg_error_exit_handler;

	/* More stuff */
	FILE * outfile;		/* target file */
	jpg::JSAMPROW row_pointer[1];	/* pointer to JSAMPLE row[s] */
	jpg::JSAMPROW line_buffer = new unsigned char [3*width] ;

	/* Step 1: allocate and initialize JPEG compression object */

	/* We have to set up the error handler first, in case the initialization
	* step fails.  (Unlikely, but it could happen if you are out of memory.)
	* This routine fills in the contents of struct jerr, and returns jerr's
	* address which we place into the link field in cinfo.
	*/
	cinfo.err = jpg::jpeg_std_error(&jerr);
	/* Now we can initialize the JPEG compression object. */
	/* Note: the original example code used the jpeg_create_compress() macro,
	 * but this can not be used since it doesn't handle the jpg namespace correctly. */
	jpg::jpeg_CreateCompress(&cinfo, JPEG_LIB_VERSION, \
			(size_t) sizeof(struct jpg::jpeg_compress_struct)) ;

	/* Step 2: specify data destination (eg, a file) */
	/* Note: steps 2 and 3 can be done in either order. */

	/* Here we use the library-supplied code to send compressed data to a
	* stdio stream.  You can also write your own code to do something else.
	* VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
	* requires it in order to write binary files.
	*/

	if ((outfile = fopen ((const char*)path, "wb")) == NULL)
	{
		const char* errstr = strerror (errno) ;
		CString msg ;
		msg.Format ("Failed to open '%s' for writing.\n%s", (LPCTSTR)path, errstr) ;
		throw msg ;
	}

	jpg::jpeg_stdio_dest(&cinfo, outfile);

	/* Step 3: set parameters for compression */

	/* First we supply a description of the input image.
	* Four fields of the cinfo struct must be filled in:
	*/
	cinfo.image_width = width ; 	/* image width and height, in pixels */
	cinfo.image_height = height ;
	cinfo.input_components = 3 ;		/* # of color components per pixel */
	cinfo.in_color_space = jpg::JCS_RGB ; 	/* colorspace of input image */

	/* Now use the library's routine to set default compression parameters.
	* (You must set at least cinfo.in_color_space before calling this,
	* since the defaults depend on the source color space.)
	*/
	jpg::jpeg_set_defaults(&cinfo);
	/* Now you can set any non-default parameters you wish to.
	* Here we just illustrate the use of quality (quantization table) scaling:
	*/
	jpg::jpeg_set_quality(&cinfo, m_quality, TRUE /* limit to baseline-JPEG values */);

	/* Step 4: Start compressor */

	/* TRUE ensures that we will write a complete interchange-JPEG file.
	* Pass TRUE unless you are very sure of what you're doing.
	*/
	jpg::jpeg_start_compress(&cinfo, TRUE);

	/* Step 5: while (scan lines remain to be written) */
	/*           jpeg_write_scanlines(...); */

	/* Here we use the library's state variable cinfo.next_scanline as the
	* loop counter, so that we don't have to keep track ourselves.
	* To keep things simple, we pass one scanline per call; you can pass
	* more if you wish, though.
	*/

	row_pointer[0] = line_buffer ;
	while (cinfo.next_scanline < cinfo.image_height)
	{
		/* jpeg_write_scanlines expects an array of pointers to scanlines.
		 * Here the array is only one element long, but you could pass
		 * more than one scanline at a time if that's more convenient.
		 */
		// Copy RGB to line_buffer.
		const rgbcolour_t* rgb_line = rgb_data + (cinfo.image_height-1 - cinfo.next_scanline)*width ;
		for (unsigned int i = 0 ; i < cinfo.image_width ; i++)
		{
			rgbcolour_t col = rgb_line[i] ;
			line_buffer[3*i+0] = (unsigned char)((col >> 16) & 0xFF) ;	// Red
			line_buffer[3*i+1] = (unsigned char)((col >>  8) & 0xFF) ;	// Green
			line_buffer[3*i+2] = (unsigned char)(col & 0xFF) ;	// Blue
		}
		(void)jpg::jpeg_write_scanlines(&cinfo, row_pointer, 1);
	}


	/* Step 6: Finish compression */

	jpg::jpeg_finish_compress(&cinfo);
	/* After finish_compress, we can close the output file. */
	fclose(outfile);


	/* Step 7: release JPEG compression object */

	/* This is an important step since it will release a good deal of memory. */
	jpg::jpeg_destroy_compress(&cinfo);

	/* And we're done! */

	delete[] line_buffer ;
}


/*########################################################################
	End
########################################################################*/
