/*########################################################################
	File:			Palette.cpp
	Creation date:	25th May 2002

	Copyright 2002 Mark Williams, all rights reserved.
	Contact: mark@extrabit.com
	http://www.extrabit.com

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

########################################################################*/

/*########################################################################
	Headers
########################################################################*/

#include "StdAfx.h"

#include <math.h>

#include "Palette.h"

/*########################################################################
	Local declarations
########################################################################*/

/*########################################################################
	Local definitions
########################################################################*/

static const double GAMMA = 2.2 ;
static const double INV_GAMMA = 1.0/GAMMA ;


/**	Constructor.
 */
Palette::Palette (int ncolours)
{
	m_type = PALETTE_BRIGHT ;
	m_main_colour = 0 ;			// Black.
	init (ncolours) ;
}


/**	Destructor.
 */
Palette::~Palette (void)
{
	delete[] m_colours ;
	delete[] m_linear_colours ;
}


/**	Resizes this Palette to contain the given number of colours.
 */
void Palette::resize (int ncolours)
{
	delete[] m_colours ;
	delete[] m_linear_colours ;
	init (ncolours) ;
}


/**	Randomly sets the colours in the palette, except the
 *	last colour which is set to black.
 */
void Palette::randomise (void)
{
	switch (m_type)
	{
	case PALETTE_BRIGHT: randomise_bright() ; break ;
	case PALETTE_PASTEL: randomise_pastel() ; break ;
	case PALETTE_MONOCHROME: randomise_monochrome() ; break ;
	case PALETTE_WHITE: set_white() ; break ;
	}
	update_linear_colours() ;
}


/**	Sets the last colour in the palette, which is used for pixels that have
 *	reached the maximum number of iterations without diverging.
 */
void Palette::set_main_colour (rgbcolour_t col)
{
	m_main_colour = col ;
	m_colours[m_ncolours-1] = col ;
	update_linear_colours() ;
}


/**	@return Returns the gamma corrected colour given the linear colour.
 */
double Palette::gamma_correct (double linear_col) const
{
	return pow (linear_col, INV_GAMMA) ;
}


/**	Initialises this Palette to the given size.
 */
void Palette::init (int ncolours)
{
	m_ncolours = ncolours ;
	m_colours = new rgbcolour_t [m_ncolours] ;
	m_linear_colours = new float[3*m_ncolours] ;

	for (int i = 0 ; i < m_ncolours ; i++)
	{
		m_colours[i] = 0 ;
	}
}


/**	Fills the palette with randomly selected bright colours.
 */
void Palette::randomise_bright (void)
{
	for (int i = 0 ; i < m_ncolours-1 ; i++)
	{
		unsigned int r = rand() & 0xFF ;
		unsigned int g = rand() & 0xFF ;
		unsigned int b = rand() & 0xFF ;

		// Favour primary colours.
		// This is done by finding the brightest colour component, then
		// halving the other components.
		if (r > g)
		{
			g /= 2 ;
			if (r > b)
			{
				// red max.
				b /= 2 ;
			}
			else
			{
				// blue max.
				r /= 2 ;
			}
		}
		else
		{
			r /= 2 ;
			if (g > b)
			{
				// green max.
				b /= 2 ;
			}
			else
			{
				// blue max.
				g /= 2 ;
			}
		}

		m_colours[i] = (r << 16) | (g << 8) | b ;
	}

	m_colours[i] = m_main_colour ;
}


/**	Fills the palette with randomly selected pastel colours.
 */
void Palette::randomise_pastel (void)
{
	for (int i = 0 ; i < m_ncolours-1 ; i++)
	{
		unsigned int r = rand() & 0xFF ;
		unsigned int g = rand() & 0xFF ;
		unsigned int b = rand() & 0xFF ;

		m_colours[i] = (r << 16) | (g << 8) | b ;
	}

	m_colours[i] = m_main_colour ;
}


/**	Fills the palette with randomly selected shades of gray.
 */
void Palette::randomise_monochrome (void)
{
	for (int i = 0 ; i < m_ncolours-1 ; i++)
	{
		unsigned int y = rand() & 0xFF ;
		m_colours[i] = (y << 16) | (y << 8) | y ;
	}

	m_colours[i] = m_main_colour ;
}


/**	Sets the palette to all white, except the last entry which is set
 *	to the main colour.
 */
void Palette::set_white (void)
{
	for (int i = 0 ; i < m_ncolours-1 ; i++)
	{
		m_colours[i] = 0x00FFFFFF ;
	}

	m_colours[i] = m_main_colour ;
}


/**	This is called whenever the colours in the palette are changed.
 */
void Palette::update_linear_colours (void)
{
	int j = 0 ;
	for (int i = 0 ; i < m_ncolours ; i++)
	{
		double r = (double)((m_colours[i] >> 16) & 0xFF) / 255.0 ;
		double g = (double)((m_colours[i] >>  8) & 0xFF) / 255.0 ;
		double b = (double)((m_colours[i]      ) & 0xFF) / 255.0 ;
		r = pow (r, GAMMA) ;
		g = pow (g, GAMMA) ;
		b = pow (b, GAMMA) ;
		m_linear_colours[j++] = (float)r ;
		m_linear_colours[j++] = (float)g ;
		m_linear_colours[j++] = (float)b ;
	}
}


/*########################################################################
	End
########################################################################*/
