/*########################################################################
	File:			ProgressiveImageGen.cpp
	Creation date:	6th June 2002

	Copyright 2002 Mark Williams, all rights reserved.
	Contact: mark@extrabit.com
	http://www.extrabit.com

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

########################################################################*/

/*########################################################################
	Headers
########################################################################*/

#include "StdAfx.h"

#include "Processor.h"
#include "Palette.h"
#include "FractalImage.h"
#include "ProgressiveImageGen.h"

/*########################################################################
	Local declarations
########################################################################*/

/*########################################################################
	Local definitions
########################################################################*/


/**	Constructor.
 *	@param image The image that this ProgressiveImageGen will generate.
 */
ProgressiveImageGen::ProgressiveImageGen (FractalImage* image)
{
	m_image = image ;
	m_pixel_values = 0 ;
	m_pixels_done = 0 ;
	m_lines_done = 0 ;
	m_current_itn = 0 ;
	m_done = false ;
	m_nextline = 0 ;
	m_curr_itn_all_done = false ;
	m_curr_itn_any_changed = false ;
}


/**	Destructor.
 */
ProgressiveImageGen::~ProgressiveImageGen (void)
{
	delete[] m_pixel_values ;
	delete[] m_pixels_done ;
	delete[] m_lines_done ;
}


/**	Overriden from ImageGen base class.
 *	Initialises thie ProgressiveImageGen for processing the current FractalImage.
 *	This must be called before any processing is done.
 *	This must also be called if the size of the image ever changes.
 */
void ProgressiveImageGen::init (void)
{
	delete[] m_pixel_values ;
	delete[] m_pixels_done ;
	delete[] m_lines_done ;

	int width  = m_image->inq_width() ;
	int height = m_image->inq_height() ;

	m_pixel_values = new ComplexNumber [width*height] ;
	m_pixels_done  = new bool [width*height] ;
	m_lines_done   = new bool [height] ;
}


/**	Overriden from ImageGen base class.
 *	Restarts the processing.
 *	This should be called whenever any attribute that affects the image
 *	is changed, such as the view transform or the palette.
 */
void ProgressiveImageGen::restart (void)
{
	// Called whenever processing needs to restart
	// i.e. due to image size or view transform being changed,
	// or fractal parameters being changed.

	// Calculate 'm_xoffs' and 'm_yoffs'.
	const Matrix33& view_transform = m_image->inq_view_transform() ;
	Vector2 org   (0.0, 0.0) ;
	Vector2 org_x (1.0, 0.0) ;
	Vector2 org_y (0.0, 1.0) ;
	view_transform.transform (&org) ;
	view_transform.transform (&org_x) ;
	view_transform.transform (&org_y) ;
	m_xoffs = org_x - org ;
	m_yoffs = org_y - org ;

	// Set 'm_origin' to the centre of the first pixel in the first (bottom) row.
	m_origin = Vector2 (0.5, 0.5) ;
	view_transform.transform (&m_origin) ;

	// Initialise the pixels.
	init_pixels() ;

	m_rowstart = m_origin ;
	m_current_itn = 0 ;
	m_done = false ;
	m_nextline = 0 ;
	m_curr_itn_all_done = true ;		// Assume all done until any are found that are not.
	m_curr_itn_any_changed = false ;
}


/**	Overriden from ImageGen base class.
 *	Processes the current line.
 *	@return Returns true if the display should be updated, otherwise false.
 */
bool ProgressiveImageGen::process (void)
{
	bool update_display = false ;
	if (!m_lines_done[m_nextline])
	{
		process_line() ;
	}

	m_rowstart += m_yoffs ;
	m_nextline++ ;
	if (m_nextline == m_image->inq_height())
	{
		// Just finished an iteration.
		m_current_itn++ ;
		int max_itns = m_image->inq_palette()->inq_ncolours()-1 ;
		if (m_curr_itn_all_done || m_current_itn == max_itns)
		{
			// No more unfinished pixels.
			m_done = true ;
		}
		if (m_curr_itn_any_changed)
		{
			update_display = true ;
		}
		m_rowstart = m_origin ;
		m_nextline = 0 ;
		m_curr_itn_all_done = true ;		// Assume all done until any are found that are not.
		m_curr_itn_any_changed = false ;
	}

	return update_display ;
}


/**	Overriden from ImageGen base class.
 *	@return Returns true when image has been generated.
 */
bool ProgressiveImageGen::inq_done (void) const
{
	return m_done ;
}


/**	Overriden from ImageGen base class.
 *	@return Returns the value shown as a progress indicator in the status bar.
 */
int ProgressiveImageGen::inq_progress (void) const
{
	return m_current_itn ;
}


/**	Initialises the FractalPixels for generating a new image.
 */
void ProgressiveImageGen::init_pixels (void)
{
	ComplexNumber initial_z = m_image->inq_initial_z() ;

	int n = m_image->inq_npixels() ;
	for (int i = 0 ; i < n ; i++)
	{
		m_pixel_values[i] = initial_z ;
		m_pixels_done[i] = false ;
	}

	int height = m_image->inq_height() ;
	for ( i = 0 ; i < height ; i++)
	{
		m_lines_done[i] = false ;
	}
}


/**	Performs one iteration on all the pixels on the given line.
 *	@return Returns a status value which has PROC_ALL_DONE set if all
 *	pixels on this line are fully computed, and PROC_CHANGED if any pixels
 *	have just been fully computed which means the display needs to be updated.
 */
void ProgressiveImageGen::process_line (void)
{
	int width = m_image->inq_width() ;
	ComplexNumber* row_pixel_values = &m_pixel_values[m_nextline*width] ;
	bool* row_pixels_done = &m_pixels_done[m_nextline*width] ;
	unsigned short* row_pixel_itns = m_image->inq_pixel_itns() + m_nextline*width ;
	rgbcolour_t* row_rgb_pixels = m_image->inq_rgb_pixels() + m_nextline*width ;
	const rgbcolour_t* colours = m_image->inq_palette()->inq_colours() ;
	bool all_done = true ;
	Vector2 v = m_rowstart ;
	const Processor* processor = m_image->inq_processor() ;
	for (int i = 0 ; i < width ; i++)
	{
		if (!row_pixels_done[i])
		{
			ComplexNumber z (row_pixel_values[i]) ;
			ComplexNumber u = m_image->convert_point_to_z (v) ;
			z = processor->iterate (z, u) ;
			row_pixel_values[i] = z ;
			if (z.inq_magnitude2() > 4.0)
			{
				row_pixel_itns[i] = m_current_itn ;
				row_rgb_pixels[i] = colours[m_current_itn] ;
				row_pixels_done[i] = true ;
				m_curr_itn_any_changed = true ;
			}
			else
			{
				all_done = false ;
			}
		}
		v += m_xoffs ;
	}

	if (all_done)
	{
		// All pixels on this line are now done, so set a flag for this line.
		m_lines_done[m_nextline] = true ;
	}
	else
	{
		// Not all pixels on this line are done, so set flag to indicate that 
		// more processing will be needed.
		m_curr_itn_all_done = false ;
	}
}


/*########################################################################
	End
########################################################################*/
