#ifndef ProgressiveImageGen_H
#define ProgressiveImageGen_H
/*########################################################################
	File:			ProgressiveImageGen.h
	Creation date:	6th June 2002

	Copyright 2002 Mark Williams, all rights reserved.
	Contact: mark@extrabit.com
	http://www.extrabit.com

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

########################################################################*/

/*########################################################################
	Headers
########################################################################*/

#include "ImageGen.h"
#include "Vector2.h"

/*########################################################################
	Declarations
########################################################################*/

class FractalImage ;
class Palette ;
class Processor ;
class ComplexNumber ;

/*########################################################################
	Class Definition
########################################################################*/

/**	This class processes a FractalImage by iterating the pixel values progressively.
 *	This allows the display to be updated regularly, rather than having to wait
 *	before anything is seen.
 *	This class maintains the state that allows the processing to be done in
 *	many incremental steps, which are called from OnIdle().
 *	Multiple passes are run over all pixels so that the whole image is updated
 *	regularly, rather than calculating each pixel one by one.
 *	On each iteration, all pixels that have exceed the divergence threshold
 *	are set using the current iteration value.
 *	Although this approach is slower, it gives a more interesting image more quickly
 *	that develops over time.
 */
class ProgressiveImageGen : public ImageGen
{
public:
	ProgressiveImageGen (FractalImage*) ;
	~ProgressiveImageGen (void) ;

	// Overriden from ImageGen base class.
	void init (void) ;
	void restart (void) ;
	bool process (void) ;
	bool inq_done (void) const ;
	int inq_progress (void) const ;

private:
	void init_pixels (void) ;
	void process_line (void) ;

private:
	/**	FractalImage. */
	FractalImage* m_image ;

	/**	Current value of 'z' at each pixel.
	 *	This array has one element per pixel, stored in the same order as
	 *	the pixels in the FractalImage.
	 */
	ComplexNumber* m_pixel_values ;

	/**	Flag for each pixel, set true once the value of 'z' at that pixel
	 *	has exceeded the divergence threshold.
	 *	This array has one element per pixel, stored in the same order as
	 *	the pixels in the FractalImage.
	 */
	bool* m_pixels_done ;

	/**	Flags for each line, set once all pixels on that line have been computed.
	 *	This array has 'image-height' elements.
	 */
	bool* m_lines_done ;

	/**	Position in fractal space of first pixel on the first (bottom) row. */
	Vector2 m_origin ;

	/**	Position in fractal space of first pixel on the current row. */
	Vector2 m_rowstart ;

	/**	Offset in fractal space from a pixel to the next.
	 *	This is determined from the view transform and the image size.
	 */
	Vector2 m_xoffs ;

	/**	Offset in fractal space from a pixel in one row to the next.
	 *	This is determined from the view transform and the image size.
	 */
	Vector2 m_yoffs ;

	/** Current iteration - each pixel in the whole image is advanced
	 *	one iteration at a time.
	 */
	unsigned short m_current_itn ;

	/** Flag set when no more processing to be done. */
	bool m_done ;

	/** Next line to process in OnIdle. */
	int m_nextline ;

	/**	This flag is set at the beginning of each iteration, and cleared if any pixels
	 *	are found that are not complete.
	 */
	bool m_curr_itn_all_done ;

	/**	This flag is cleared at the beginning of each iteration, and set if any pixels
	 *	are completed on this iteration, which indicates that the display needs to be updated.
	 */
	bool m_curr_itn_any_changed ;
} ;


/*########################################################################
	End
########################################################################*/
#endif
